/************************************************************************/
/*                                                                      */
/*                      I2C Sniffer					*/
/*                                                                      */
/*              Author: Peter Dannegger                                 */
/*                                                                      */
/************************************************************************/
#include "main.h"
#include "uart.h"
#include "i2csniff.h"

void init_i2c( void )
{
  SCL 		= 1;
  SCL_DDR 	= 1;						// force busy waiting
  SDA 		= 1;
  SDA_DDR 	= 0;						// listen only, no ACK, no data
  USICR = 1<<USIWM1^1<<USIWM0;			// I2C, no counter
}


void putnibble( u8 j )
{
  j &= 0x0F;
  if( j > 9 )
    j += 'A' - '0' - 10;
  uart_putc( j + '0' );
}

int CheckDataMatch( u8 data_match_p[] , u8 data_len )
{
	u8 i;

	for ( i = 0 ; i < data_len ; i++ )
	{
		if (!data_match_p[i])
			return 0;
	}
	return 1;
}

void i2c_sniff( void )
{
  u8 get_ack 		= 0;
  u8 START_Received = 0;
  u8 Buffer			= 0;
  u8 DeviceAddress	= 0;
  u8 DeviceMask		= 0xFF;
  u8 DeviceMatch	= 0;
  u8 DeviceTrigger  = 0;
  u16 DataReceived  = 0;
  u8 DataTriggerLen	= 0;
  u8 len			= 0;
  u8 i				= 0;
  u8 I2C_Counter	= 0;
  u8 RS232_Buffer[9];
  u8 DataTrigger[6];
  u8 DataTriggerMatch[3];

  for(;;){

	DataReceived = uart_getc();
	if ( DataReceived & UART_NO_DATA )
	{
	    /* 
	     * no data available from UART 
	     */
	    if( USISR & 1<<USIPF ){				// STOP received
		  while( !SDA_PIN );				// wait until SDA = high
	      USICR = 1<<USIWM1^1<<USIWM0;		// I2C, no counter
	      USISR = 1<<USIPF;					// clear USIPF
	      uart_puts( STOP LINEFEED );
		  PORTD &=~ (1<<5);					// clear trigger output
	    }
	
		if( USISR & 1<<USISIF ){		// START received
	      while( SCL_PIN );						// wait until SCL = low
	      USICR = 1<<USIWM1^1<<USIWM0^1<<USICS1^1<<USICS0;
												// I2C, get data on neg. edge
	      USISR = 1<<USISIF^1<<USIOIF^(0x0F & -16);
												// clear USISIF, count 16 edges
	      get_ack = 0;
		  I2C_Counter = 0;
		  DeviceMatch = 0;
		  START_Received = 1;
		  memset(&DataTriggerMatch, 0, 3);
		  uart_putc( START );
	    } 
	
		if( USISR & 1<<USIOIF ){		// Data / ACK received
			Buffer = USIDR;
			if( ++get_ack & 1 ){				// toggle read data / ACK
				USISR = 1<<USIOIF^(0x0F & -2);	// count 2 edges
				if ( START_Received ) {	
					START_Received = 0;	
					DeviceAddress = Buffer >> 1;
										
					if (( DeviceAddress & DeviceMask ) == DeviceTrigger )
					{
						if ( DataTriggerLen == 0 )
							PORTD |= (1<<5);			// set trigger output to high

						if ( DataTriggerLen > 0 )
							DeviceMatch = 1;
					}											
							
					putnibble( DeviceAddress >> 4 );
					putnibble( DeviceAddress );
					uart_putc( (Buffer & 5) ? READ : WRITE );				
				}else{
					if (I2C_Counter < DataTriggerLen )
					{
						if (( Buffer & DataTrigger[(I2C_Counter * 2) + 1] ) == DataTrigger[I2C_Counter * 2] )
						{
							if ( DeviceMatch )
								DataTriggerMatch[I2C_Counter] = 1;								
						}
					}

					if ( CheckDataMatch(DataTriggerMatch, DataTriggerLen ) )
					{
						PORTD |= (1<<5);			// set trigger output to high
					}
							
					I2C_Counter++;
					putnibble( Buffer >> 4 );
					putnibble( Buffer );
				}
			
			}else{
				USISR = 1<<USIOIF^(0x0F & -16);	// count 16 edges
				uart_putc( (Buffer & 1) ? NACK : ACK );
			}
	    }
	} else {
	    /*
	     * new data available from UART
	     * check for Frame or Overrun error
	     */
		len = (u8) DataReceived;
		memset(&RS232_Buffer, 0, sizeof(RS232_Buffer));


		for ( i = 1 ; i < len ; i++ )
		{
			DataReceived = UART_NO_DATA;

			// wait for next byte to receive:
			while ( DataReceived & UART_NO_DATA )
				DataReceived = uart_getc();

			RS232_Buffer[i - 1] = (u8) DataReceived;

		}

		DeviceTrigger 	= RS232_Buffer[0];
		DeviceMask 		= RS232_Buffer[1]; 

		// apply device mask:
		DeviceTrigger 	&= DeviceMask;

		DataTriggerLen	= RS232_Buffer[2];
		memset(&DataTrigger, 0, sizeof(DataTrigger));
		memset(&DataTriggerMatch, 0, sizeof(DataTriggerMatch));

		for ( i = 0 ; i < DataTriggerLen ; i++ )
		{
			DataTrigger[i * 2] 			= RS232_Buffer[3 + (i * 2)];
			DataTrigger[(i * 2) + 1] 	= RS232_Buffer[4 + (i * 2)];
			// apply data mask:
			DataTrigger[i * 2] 			&= DataTrigger[(i * 2) + 1];			
		}
	} 
				
  }
}
